/*
      This program implements the algorithm given by George
W. Hart in the September 1994 edition of the Communications
of the ACM (Volume 37, Number 9).

      Edgar Allan Poe used this method to crack simple
substitution ciphers around 1839; consider the following
from pages 784-785 of David Kahn's The Codebreakers:

           The closest he ever came to doing so was
           when he demonstrated how he deduced that
           a challenge sent him by G.  W. Kulp, of
           Lewiston, Pennsylvania, was a false
           cryptogram.  He picked out three words
           in the cryptogram--MW, LAAM, and MLW.
           Since "all English words of but two
           letters consist of a vowel and a
           consonant," he wrote, MW must be one of
           30 words, which he listed. He then
           inserted every letter of the alphabet in
           the middle of all 30 words in an
           exhaustive trial process to see which
           letters would make a sensible word out
           of MLW. Here he found 18, including ash
           and tho'.  Turning to LAAM, he observed,
           that "if MLW be ash, then LAAM will be a
           word of this form, s..a, in which the
           dots represent two unknown letters of
           the same kind." He ran through his 18
           words in this way, and found that the
           only one that gave a possible meaning
           for LAAM was h..t, or hoot.  "LAAM is
           then hoot or nothing. But the hypothesis
           of the word hoot is founded upon that of
           the word tho'.... We now arrive at a
           definite conclusion.  Either Mr. Kulp's
           puzzle is not genuine, or MW stands for
           to, MLW for tho', and LAAM for hoot.
           But it is evident that this latter
           cannot be--for in that case both W and A
           represent the letter 0.  What
           follows?--why that Mr. Kulp's puzzle is
           no puzzle at all. This demonstration is
           as absolutely conclusive as any
           mathematical one could be. The process
           of reasoning here employed is that
           employed also in the solution of the
           cyphers."
*/

#include <tigcclib.h>

#define TRUE  1
#define FALSE 0

#define ALPHABET_SIZE 27 /* A-Z '; > 26 */

typedef struct CipherTextNode
          {
            char                  *szCipherText;
            struct CipherTextNode *pNext;
          } *CipherTextNodePtr;

typedef struct PlainTextNode
          {
            char                 *szPlainText;
            struct PlainTextNode *pNext;
          } *PlainTextNodePtr;

typedef struct PatternNode
          {
            int                nDistinctLetters;
            char               *szPattern;
            unsigned long      ulPlainTexts;
            PlainTextNodePtr   pPlainTextHead;
            struct PatternNode *pLesser;
            struct PatternNode *pGreater;
          } *PatternNodePtr;

typedef struct SortedNode
          {
            char              *szCipherText;
            PatternNodePtr    pPattern;
            PlainTextNodePtr  pPlainText;
            struct SortedNode *pPrevious;
            struct SortedNode *pNext;
          } *SortedNodePtr;

typedef struct EncipheredNode
          {
            char                  *szCipherText;
            PatternNodePtr        pPattern;
            SortedNodePtr         pSorted;
            struct EncipheredNode *pNext;
          } *EncipheredNodePtr;

static int            Aborted(void);
static PatternNodePtr AddEncipheredWord(char *,EncipheredNodePtr *,
                       EncipheredNodePtr *,PatternNodePtr *,int *);
static void           AddPlainText(PatternNodePtr,char *,char *,int *);
static void           AddWordNotInDictionary(PatternNodePtr,int *);
static void           FreePattern(PatternNodePtr);
static int            GetCipherChar(CipherTextNodePtr *,char **);
static void           GetPatternForPlainTextWord(char *,char *,int *);
static void           GetCipherText(CipherTextNodePtr *,int *);
       int            _main(void);
static int            MyStricmp(char *,char *);
static void           OutputSolution(EncipheredNodePtr,CipherTextNodePtr,
                       unsigned long *,char *,int *);
static PatternNodePtr PatternForEncipheredWord(char *,PatternNodePtr *,
                       int *);
static char           *PatternForWord(char *,int *,int *);
static void           Solve(int,EncipheredNodePtr,SortedNodePtr,
                       CipherTextNodePtr,int *);
/*
   Replace the following word list with your own or modify
   the program to read the list from a file.
*/
static char *g_pszWordList [] =
              {
                "a","able","about","above","according",
                "account","across","act","action",
                "activities","activity","actually",
                "added","addition","administration",
                "after","again","against","age","ago",
                "ahead","aid","air","all","almost",
                "alone","along","already","also",
                "although","always","am","america",
                "american","among","amount","an",
                "analysis","and","another","answer",
                "anti","any","anyone","anything",
                "apparently","appear","appeared",
                "approach","are","area","areas","army",
                "around","art","as","ask","asked",
                "association","at","attack","attention",
                "audience","available","average","away",
                "b","back","bad","ball","based",
                "basic","basis","be","beautiful",
                "became","because","become","bed",
                "been","before","began","beginning",
                "behind","being","believe","below",
                "best","better","between","beyond",
                "big","bill","black","blood","blue",
                "board","body","book","born","both",
                "boy","boys","bring","brought","brown",
                "building","built","business","but",
                "by","c","call","called","came","can",
                "cannot","car","care","carried","cars",
                "case","cases","cause","cent","center",
                "central","century","certain",
                "certainly","change","character",
                "chief","child","children","choice",
                "christian","church","city","class",
                "clear","clearly","close","closed",
                "club","co","cold","college","color",
                "come","comes","coming","committee",
                "common","communist","community",
                "company","complete","completely",
                "concerned","conditions","congress",
                "consider","considered","continued",
                "control","corner","corps","cost",
                "costs","could","couldn\'t","countries",
                "country","county","couple","course",
                "court","covered","cut","d","daily",
                "dark","data","day","days","de",
                "dead","deal","death","decided",
                "decision","deep","defense","degree",
                "democratic","department","described",
                "design","designed","determined",
                "developed","development","did",
                "didn\'t","difference","different",
                "difficult","direct","direction",
                "directly","distance","district","do",
                "does","doing","done","don\'t","door",
                "doubt","down","dr","drive","due",
                "during","e","each","earlier","early",
                "earth","east","economic","education",
                "effect","effective","effects","effort",
                "efforts","eight","either","elements",
                "else","end","england","english",
                "enough","entire","equipment",
                "especially","established","europe",
                "even","evening","ever","every",
                "everything","evidence","example",
                "except","existence","expect",
                "expected","experience","extent","eye",
                "eyes","f","face","fact","faith",
                "fall","family","far","farm","father",
                "fear","federal","feed","feel",
                "feeling","feet","felt","few","field",
                "figure","figures","filled","final",
                "finally","find","fine","fire","firm",
                "first","fiscal","five","floor",
                "followed","following","food","foot",
                "for","force","forces","foreign",
                "form","former","forms","forward",
                "found","four","free","freedom",
                "french","friend","friends","from",
                "front","full","function","further",
                "future","g","game","gave","general",
                "generally","george","get","getting",
                "girl","girls","give","given","gives",
                "glass","go","god","going","gone",
                "good","got","government","great",
                "greater","green","ground","group",
                "groups","growing","growth","gun","h",
                "had","hair","half","hall","hand",
                "hands","happened","hard","has","have",
                "having","he","head","hear","heard",
                "heart","heavy","held","hell","help",
                "her","here","herself","he\'s","high",
                "higher","him","himself","his",
                "history","hit","hold","home","hope",
                "horse","hospital","hot","hotel",
                "hour","hours","house","how","however",
                "human","hundred","husband","i","idea",
                "ideas","if","i\'ll","i\'m","image",
                "immediately","important","in",
                "include","including","income",
                "increase","increased","indeed",
                "individual","industrial","industry",
                "influence","information","inside",
                "instead","interest","international",
                "into","involved","is","island",
                "issue","it","its","it\'s","itself",
                "i\'ve","j","job","john","just",
                "justice","keep","kennedy","kept",
                "kind","knew","know","knowledge",
                "known","l","labor","lack","land",
                "language","large","larger","last",
                "late","later","latter","law","lay",
                "lead","leaders","learned","least",
                "leave","led","left","length","less",
                "let","letter","letters","level",
                "life","light","like","line","lines",
                "list","literature","little","live",
                "lived","living","local","long",
                "longer","look","looked","looking",
                "lost","lot","love","low","lower","m",
                "made","main","major","make","makes",
                "making","man","manner","mans","many",
                "march","market","married","mass",
                "material","matter","may","maybe","me",
                "mean","meaning","means","medical",
                "meet","meeting","member","members",
                "men","merely","met","method",
                "methods","middle","might","miles",
                "military","million","mind","minutes",
                "miss","modern","moment","money",
                "month","months","moral","more",
                "morning","most","mother","move",
                "moved","movement","moving","mr","mrs",
                "much","music","must","my","myself",
                "n","name","nation","national",
                "nations","natural","nature","near",
                "nearly","necessary","need","needed",
                "needs","neither","never",
                "new","next","night","no","non","nor",
                "normal","north","not","note",
                "nothing","now","nuclear","number",
                "numbers","obtained","obviously","of",
                "off","office","often","oh","old",
                "on","once","one","ones","one\'s",
                "only","open","opened","operation",
                "opportunity","or","order",
                "organization","other","others","our",
                "out","outside","over","own","p",
                "paid","paper","part","particular",
                "particularly","parts","party","passed",
                "past","pattern","pay","peace",
                "people","per","performance","perhaps",
                "period","person","personal","persons",
                "physical","picture","piece","place",
                "plan","plane","planning","plans",
                "plant","play","point","points",
                "police","policy","political","pool",
                "poor","population","position",
                "possible","post","power","present",
                "president","press","pressure","price",
                "principle","private","probably",
                "problem","problems","process",
                "production","products","program",
                "programs","progress","property",
                "provide","provided","public","purpose",
                "put","quality","question","questions",
                "quite","r","race","radio","ran",
                "range","rate","rather","reached",
                "reaction","read","reading","ready",
                "real","really","reason","received",
                "recent","recently","record","red",
                "religion","religious","remember",
                "report","reported","required",
                "research","respect","responsibility",
                "rest","result","results","return",
                "returned","right","river","road",
                "room","run","running","s","said",
                "sales","same","sat","saw","say",
                "saying","says","school","schools",
                "science","season","second","secretary",
                "section","see","seem","seemed",
                "seems","seen","self","sense","sent",
                "series","serious","served","service",
                "services","set","seven","several",
                "shall","she","short","shot","should",
                "show","showed","shown","side",
                "similar","simple","simply","since",
                "single","situation","six","size",
                "slowly","small","so","social",
                "society","some","something",
                "sometimes","somewhat","son","soon",
                "sort","sound","south","southern",
                "soviet","space","speak","special",
                "specific","spirit","spring","square",
                "st","staff","stage","stand",
                "standard","start","started","state",
                "statements","states","stay","step",
                "steps","still","stock","stood","stop",
                "stopped","story","straight","street",
                "strength","strong","student",
                "students","study","subject","such",
                "suddenly","summer","sun","support",
                "sure","surface","system","t","table",
                "take","taken","taking","talk","tax",
                "technical","tell","temperature","ten",
                "term","terms","test","th","than",
                "that","thats","the","their","them",
                "themselves","then","theory","there",
                "therefore","there\'s","these","they",
                "thing","things","think","thinking",
                "third","thirty","this","those",
                "thought","three","through",
                "throughout","thus","time","times",
                "to","today","together","told","too",
                "took","top","total","toward","town",
                "trade","training","trial","tried",
                "trouble","true","truth","try",
                "trying","turn","turned","twenty",
                "two","type","types","u","unlikely",
                "under","understand","union","united",
                "university","until","up","upon","us",
                "use","used","using","usually","value",
                "values","various","very","view",
                "voice","volume","waiting","walked",
                "wall","want","wanted","war","was",
                "washington","wasn\'t","water","way",
                "ways","we","week","weeks","well",
                "went","were","west","western","what",
                "whatever","when","where","whether",
                "which","while","white","who","whole",
                "whom","whose","why","wide","wife",
                "will","william","window","wish",
                "with","within","without","woman",
                "women","word","words","work","worked",
                "working","works","world","would",
                "wouldn\'t","written","wrong","wrote",
                "year","years","yes","yet","york",
                "you","young","your","you\'re",NULL
              };

static int MyStricmp(
 char *sz1,
 char *sz2)
  {
    register int result;

    while ((! (result=(tolower((int) *sz1)-tolower((int) *sz2))))
    &&     *sz1++)
      ++sz2;
    return result;
  }

static int Aborted(
 void)
  {
    int nResult;

    if (kbhit())
      {
        while (kbhit())
          ngetchx();
        nResult=TRUE;
        printf("Aborting...\n");
      }
    else
      nResult=FALSE;
    return nResult;
  }

static void GetCipherText(
 CipherTextNodePtr *ppCipherTextHead,
 int               *bErr)
  {
    size_t            nLen;
    CipherTextNodePtr pCipherText;
    CipherTextNodePtr pCipherTextTail;
    char              szLine [256];

    pCipherTextTail=NULL;
    printf("Enter cipher text followed by a blank line.\n");
    nLen=256;
    while ((! *bErr) && nLen && getsn(&szLine[0],256))
      {
        putchar((int) '\n');
        nLen=strlen(&szLine[0]);
        if (nLen)
          {
            if (szLine[nLen-1] == '\n')
              szLine[--nLen]='\0';
          }
        if (nLen)
          {
            if (szLine[nLen-1] == '\r')
              szLine[--nLen]='\0';
          }
        if (nLen)
          {
            if (pCipherText=(CipherTextNodePtr)
             malloc(sizeof(struct CipherTextNode)))
              if (pCipherText->szCipherText=(char *)
               malloc(nLen+1))
                {
                  strcpy(pCipherText->szCipherText,&szLine[0]);
                  pCipherText->pNext=NULL;
                  if (pCipherTextTail)
                    pCipherTextTail->pNext=pCipherText;
                  else
                    *ppCipherTextHead=pCipherText;
                  pCipherTextTail=pCipherText;
                }
              else
                {
                  free((void *) pCipherText);
                  *bErr=TRUE;
                  printf("Fatal error:  memory was exhausted.\n");
                }
            else
              {
                *bErr=TRUE;
                printf("Fatal error:  memory was exhausted.\n");
              }
          }
      }
    if (! *bErr)
      {
        if (*ppCipherTextHead == NULL)
          {
            *bErr=TRUE;
            printf("No cipher text was entered.\n");
          }
      }
    return;
  }

static int GetCipherChar(
 CipherTextNodePtr *ppCipherText,
 char              **ppcCipherText)
  {
    int               nResult;
    char              *pc;
    CipherTextNodePtr pCipherText;

    pCipherText=*ppCipherText;
    pc=*ppcCipherText;
    if (pCipherText)
      {
        if (*pc)
          {
            nResult=(int) *pc;
            ++pc;
          }
        else
          {
            nResult=(int) '\n';
            pCipherText=pCipherText->pNext;
            if (pCipherText)
              pc=pCipherText->szCipherText;
            else
              pc=NULL;
          }
      }
    else
      nResult=EOF;
    *ppCipherText=pCipherText;
    *ppcCipherText=pc;
    return nResult;
  }

static void OutputSolution(
 EncipheredNodePtr pEncipheredHead,
 CipherTextNodePtr pCipherTextHead,
 unsigned long     *nPossibleSolutions,
 char              *szDecipherment,
 int               *nDeciphermentCount)
  {
    int               nCurrentChar;
    int               nWordLen;
    char              *pc;
    char              *pcCipherText;
    CipherTextNodePtr pCipherText;
    EncipheredNodePtr pEnciphered;


    pCipherText=pCipherTextHead;
    pcCipherText=pCipherTextHead->szCipherText;
    ++(*nPossibleSolutions);
    pEnciphered=pEncipheredHead;
    do
      {
        while (((nCurrentChar=GetCipherChar(&pCipherText,&pcCipherText)) != EOF)
        &&      (nCurrentChar != (int) '\'')
        &&      (! islower(tolower(nCurrentChar))))
          putchar(nCurrentChar);
        if (nCurrentChar != EOF)
          {
            nWordLen=0;
            pc=pEnciphered->pSorted->pPlainText->szPlainText;
            if (*pc)
              do
                {
                  if (nCurrentChar == '\'')
                    putchar(nCurrentChar);
                  else
                    if (islower(nCurrentChar))
                      {
                        nCurrentChar=tolower((int) *pc);
                        putchar(nCurrentChar);
                      }
                    else
                      {
                        nCurrentChar=toupper((int) *pc);
                        putchar(nCurrentChar);
                      }
                  ++pc;
                }
              while (((nCurrentChar
                        =GetCipherChar(&pCipherText,&pcCipherText)) != EOF)
              &&     (nWordLen < 255)
              &&     ((nCurrentChar == '\'')
                   || (islower(tolower(nCurrentChar)))));
            else
              do
                {
                  if (nCurrentChar == '\'')
                    putchar(nCurrentChar);
                  else
                    if (islower(nCurrentChar))
                      {
                        if (nDeciphermentCount[nCurrentChar-(int) 'a'])
                          nCurrentChar=tolower((int)
                           (szDecipherment[nCurrentChar-(int) 'a']));
                        else
                          nCurrentChar=(int) '*';
                        putchar(nCurrentChar);
                      }
                    else
                      if (isupper(nCurrentChar))
                        {
                          if (nDeciphermentCount[
                           nCurrentChar-(int) 'A'])
                            nCurrentChar=toupper((int)
                             (szDecipherment[nCurrentChar-(int) 'A']));
                          else
                            nCurrentChar=(int) '*';
                          putchar(nCurrentChar);
                        }
                      else
                        {
                          nCurrentChar=toupper((int) *pc);
                          putchar(nCurrentChar);
                        }
                }
              while (((nCurrentChar
                        =GetCipherChar(&pCipherText,&pcCipherText)) != EOF)
              &&     (nWordLen < 255)
              &&     ((nCurrentChar == '\'')
                   || (islower(tolower(nCurrentChar)))));
            putchar(nCurrentChar);
            pEnciphered=pEnciphered->pNext;
          }
      }
    while (nCurrentChar != EOF);
    putchar((int) '\n');
    return;
  }

static void Solve(
 int               nDistinctWords,
 EncipheredNodePtr pEncipheredHead,
 SortedNodePtr     pSortedHead,
 CipherTextNodePtr pCipherTextHead,
 int               *bAborted)
  {
    int           bBorrow;
    int           bContradiction;
    char          cDeciphered;
    char          cEnciphered;
    int           nCurrentChar;
    int           nDeciphermentCount [26];
    int           nEnciphermentCount [26];
    int           nScore;
    int           nScoreMax;
    int           nWord;
    char          *pc1;
    char          *pc2;
    SortedNodePtr pSorted;
    char          szDecipherment [26];
    char          szEncipherment [26];
    unsigned long ulPossibleSolutions;

    ulPossibleSolutions=(unsigned long) 0;
    for (nCurrentChar=26; nCurrentChar--;)
      {
        szDecipherment[nCurrentChar]='\0';
        nDeciphermentCount[nCurrentChar]=0;
        szEncipherment[nCurrentChar]='\0';
        nEnciphermentCount[nCurrentChar]=0;
      }
    nScore=0;
    nScoreMax=0;
    pSorted=pSortedHead;
    nWord=0;
    while ((! *bAborted) && pSorted)
      {
        if (*(pSorted->pPlainText->szPlainText))
          ++nScore;
        ++nWord;
        pc1=pSorted->pPlainText->szPlainText;
        bContradiction=FALSE;
        if (*pc1) /* word in dictionary */
          {
            pc2=pSorted->szCipherText;
            while ((! bContradiction) && (*pc2))
              {
                if (*pc1 != '\'')
                  {
                    cDeciphered=(char) tolower((int) *pc1);
                    cEnciphered=(char) tolower((int) *pc2);
                    if (nDeciphermentCount[cEnciphered-'a'])
                      if (szDecipherment[cEnciphered-'a'] == cDeciphered)
                        ++(nDeciphermentCount[cEnciphered-'a']);
                      else
                        bContradiction=TRUE;
                    else
                      {
                        szDecipherment[cEnciphered-'a']=cDeciphered;
                        nDeciphermentCount[cEnciphered-'a']=1;
                      }
                    if (nEnciphermentCount[cDeciphered-'a'])
                      if (szEncipherment[cDeciphered-'a'] == cEnciphered)
                        ++(nEnciphermentCount[cDeciphered-'a']);
                      else
                        bContradiction=TRUE;
                    else
                      {
                        szEncipherment[cDeciphered-'a']=cEnciphered;
                        nEnciphermentCount[cDeciphered-'a']=1;
                      }
                  }
                ++pc1;
                ++pc2;
              }
          }
        else /* word not in dictionary */
          {
            if (nScore+nDistinctWords-nWord < nScoreMax)
              bContradiction=TRUE;
          }
        if (bContradiction)
          {
            if (*(pSorted->pPlainText->szPlainText))
              --nScore;
            --nWord;
            bBorrow=TRUE;
            while ((bBorrow) && (pSorted))
              {
                pc1=pSorted->pPlainText->szPlainText;
                if (*pc1)
                  {
                    bContradiction=FALSE;
                    pc2=pSorted->szCipherText;
                    while ((! bContradiction) && (*pc2))
                      {
                        if (*pc1 != '\'')
                          {
                            cDeciphered=(char) tolower((int) *pc1);
                            cEnciphered=(char) tolower((int) *pc2);
                            if (szDecipherment[cEnciphered-'a'] == cDeciphered)
                              {
                               if (--(nDeciphermentCount[cEnciphered-'a'])
                                == 0)
                                 szDecipherment[cEnciphered-'a']='\0';
                              }
                            else
                              bContradiction=TRUE;
                            if (szEncipherment[cDeciphered-'a'] == cEnciphered)
                              {
                                if (--(nEnciphermentCount[cDeciphered-'a'])
                                 == 0)
                                  szEncipherment[cDeciphered-'a']='\0';
                              }
                            else
                              bContradiction=TRUE;
                          }
                        ++pc1;
                        ++pc2;
                      }
                  }
                pSorted->pPlainText=pSorted->pPlainText->pNext;
                if (pSorted->pPlainText)
                  bBorrow=FALSE;
                else
                  {
                    pSorted->pPlainText=pSorted->pPattern->pPlainTextHead;
                    pSorted=pSorted->pPrevious;
                    if (pSorted)
                      {
                        if (*(pSorted->pPlainText->szPlainText))
                          --nScore;
                      }
                    --nWord;
                  }
              }
          }
        else
          if (pSorted->pNext)
            pSorted=pSorted->pNext;
          else
            {
              if (nScore > nScoreMax)
                {
                  nScoreMax=nScore;
                  clrscr();
                  ulPossibleSolutions=(unsigned long) 0;
                }
              OutputSolution(pEncipheredHead,pCipherTextHead,
               &ulPossibleSolutions,&szDecipherment[0],
               &nDeciphermentCount[0]);
              if (*(pSorted->pPlainText->szPlainText))
                --nScore;
              --nWord;
              bBorrow=TRUE;
              while ((bBorrow) && (pSorted))
                {
                  pc1=pSorted->pPlainText->szPlainText;
                  if (*pc1)
                    {
                      bContradiction=FALSE;
                      pc2=pSorted->szCipherText;
                      while ((! bContradiction) && (*pc2))
                        {
                          if (*pc1 != '\'')
                            {
                              cDeciphered=(char) tolower((int) *pc1);
                              cEnciphered=(char) tolower((int) *pc2);
                              if (szDecipherment[cEnciphered-'a']
                               == cDeciphered)
                                {
                                 if (--(nDeciphermentCount[cEnciphered-'a'])
                                  == 0)
                                   szDecipherment[cEnciphered-'a']='\0';
                                }
                              else
                                bContradiction=TRUE;
                              if (szEncipherment[cDeciphered-'a']
                               == cEnciphered)
                                {
                                  if (--(nEnciphermentCount[cDeciphered-'a'])
                                   == 0)
                                    szEncipherment[cDeciphered-'a']='\0';
                                }
                              else
                                bContradiction=TRUE;
                            }
                          ++pc1;
                          ++pc2;
                        }
                    }
                  pSorted->pPlainText=pSorted->pPlainText->pNext;
                  if (pSorted->pPlainText)
                    bBorrow=FALSE;
                  else
                    {
                      pSorted->pPlainText=pSorted->pPattern->pPlainTextHead;
                      pSorted=pSorted->pPrevious;
                      if (pSorted)
                        {
                          if (*(pSorted->pPlainText->szPlainText))
                            --nScore;
                        }
                      --nWord;
                    }
                }
            }
        *bAborted=Aborted();
      }
    if (! *bAborted)
      {
        if (ulPossibleSolutions == (unsigned long) 0)
          printf("No solutions were found.\n");
      }
    return;
  }

static void AddWordNotInDictionary(
 PatternNodePtr pPattern,
 int            *bErr)
  {
    PlainTextNodePtr pPlainText;

    if (pPlainText=(PlainTextNodePtr) malloc(sizeof(struct PlainTextNode)))
      if (pPlainText->szPlainText=(char *) malloc(sizeof(char)))
        {
          *(pPlainText->szPlainText)='\0';
          pPlainText->pNext=NULL;
          pPattern->pPlainTextHead=pPlainText;
        }
      else
        {
          *bErr=TRUE;
          printf("Fatal error:  out of memory.\n");
          free((void *) pPlainText);
        }
    else
      {
        *bErr=TRUE;
        printf("Fatal error:  out of memory.\n");
      }
    return;
  }

static void AddPlainText(
 PatternNodePtr pPatternHead,
 char           *szPattern,
 char           *szPlainText,
 int            *bErr)
   {
     int              nRelation;
     int              bFinished;
     PatternNodePtr   pInternal;
     PlainTextNodePtr pPlainText;

     pInternal=pPatternHead;
     bFinished=FALSE;
     while (! bFinished)
       {
         nRelation=strcmp(szPattern,pInternal->szPattern);
         if (nRelation < 0)
           if (pInternal->pLesser)
             pInternal=pInternal->pLesser;
           else
             bFinished=TRUE;
         else
           if (nRelation > 0)
             if (pInternal->pGreater)
               pInternal=pInternal->pGreater;
             else
               bFinished=TRUE;
           else
             {
               if (pPlainText=(PlainTextNodePtr)
                malloc(sizeof(struct PlainTextNode)))
                 if (pPlainText->szPlainText=(char *)
                  malloc(1+strlen(szPlainText)))
                   {
                     strcpy(pPlainText->szPlainText,szPlainText);
                     ++(pInternal->ulPlainTexts);
                     pPlainText->pNext=pInternal->pPlainTextHead;
                     pInternal->pPlainTextHead=pPlainText;
                   }
                 else
                   {
                     *bErr=TRUE;
                     printf("Fatal error:  out of memory.\n");
                     free((void *) pPlainText);
                   }
               else
                 {
                   *bErr=TRUE;
                   printf("Fatal error:  out of memory.\n");
                 }
               bFinished=TRUE;
             }
       }
     return;
   }

static void GetPatternForPlainTextWord(
 char *szWord,
 char *szPattern,
 int  *bAlphabetic)
  {
    char cDistinctChars;
    char *pc1;
    char *pc2;
    char *pc3;
    char *pc4;

    *bAlphabetic=TRUE;
    pc1=szWord;
    pc2=szPattern;
    while ((*bAlphabetic) && (*pc1) && (*pc1 != '\n'))
      {
        if (*pc1 != '\'')
          {
            *pc1=(char) tolower((int) *pc1);
            *bAlphabetic=islower((int) *pc1);
          }
        *pc2='\0';
        ++pc2;
        ++pc1;
      }
    if (*bAlphabetic)
      {
        *pc1='\0';
        *pc2='\0';
        cDistinctChars='\0';
        pc1=szWord;
        pc2=szPattern;
        while ((*pc1) && (*pc1 != '\n'))
          {
            if (*pc2 == '\0')
              {
                if (*pc1 == '\'')
                  *pc2=ALPHABET_SIZE;
                else
                  {
                    *pc2=++cDistinctChars;
                    pc3=pc1+1;
                    pc4=pc2+1;
                    while (*pc3)
                      {
                        if (*pc3 == *pc1)
                          *pc4=cDistinctChars;
                        ++pc4;
                        ++pc3;
                      }
                  }
              }
            ++pc2;
            ++pc1;
          }
      }
    return;
  }

static void FreePattern(
 PatternNodePtr pPattern)
  {
    PlainTextNodePtr pPlainText;

    if (pPattern)
      {
        FreePattern(pPattern->pLesser);
        FreePattern(pPattern->pGreater);
        free((void *) (pPattern->szPattern));
        while (pPattern->pPlainTextHead)
          {
            pPlainText=pPattern->pPlainTextHead->pNext;
            free((void *) (pPattern->pPlainTextHead->szPlainText));
            free((void *) (pPattern->pPlainTextHead));
            pPattern->pPlainTextHead=pPlainText;
          }
        free((void *) pPattern);
      }
    return;
  }

static char *PatternForWord(
 char *szWord,
 int  *nDistinctLetters,
 int  *bErr)
  {
    char cDistinctChars;
    char *pc1;
    char *pc2;
    char *pc3;
    char *pc4;
    char *szPattern;

    if (szPattern=(char *) malloc(1+strlen(szWord)))
      {
        pc1=szWord;
        pc2=szPattern;
        while (*pc1)
          {
            *pc2='\0';
            ++pc2;
            ++pc1;
          }
        *pc2='\0';
        cDistinctChars='\0';
        pc1=szWord;
        pc2=szPattern;
        while (*pc1)
          {
            if (*pc2 == '\0')
              {
                if (*pc1 == '\'')
                  *pc2=ALPHABET_SIZE;
                else
                  {
                    *pc2=++cDistinctChars;
                    pc3=pc1+1;
                    pc4=pc2+1;
                    while (*pc3)
                      {
                        if (*pc3 == *pc1)
                          *pc4=cDistinctChars;
                        ++pc4;
                        ++pc3;
                      }
                  }
              }
            ++pc2;
            ++pc1;
          }
        *nDistinctLetters=cDistinctChars;
      }
    else
      {
        *bErr=TRUE;
        printf("Fatal error:  out of memory.\n");
        *nDistinctLetters=0;
      }
    return szPattern;
  }

static PatternNodePtr PatternForEncipheredWord(
 char           *szEncipheredWord,
 PatternNodePtr *pPatternHead,
 int            *bErr)
  {
    int            bFinished;
    int            nDistinctLetters;
    int            nRelation;
    PatternNodePtr pExternal;
    PatternNodePtr pInternal;
    char           *szPattern;

    pInternal=NULL;
    szPattern=PatternForWord(szEncipheredWord,&nDistinctLetters,bErr);
    if (! *bErr)
      {
        if (*pPatternHead)
          {
            pInternal=*pPatternHead;
            bFinished=FALSE;
            do
              {
                nRelation=strcmp(szPattern,pInternal->szPattern);
                if (nRelation < 0)
                  if (pInternal->pLesser)
                    pInternal=pInternal->pLesser;
                  else
                    {
                      if (pExternal=(PatternNodePtr)
                       malloc(sizeof(struct PatternNode)))
                        {
                          pExternal->szPattern=szPattern;
                          pExternal->nDistinctLetters=nDistinctLetters;
                          pExternal->ulPlainTexts=(unsigned long) 0;
                          pExternal->pPlainTextHead=NULL;
                          pExternal->pLesser=NULL;
                          pExternal->pGreater=NULL;
                          pInternal->pLesser=pExternal;
                        }
                      else
                        {
                          *bErr=TRUE;
                          printf("Fatal error:  out of memory.\n");
                        }
                      pInternal=pExternal;
                      bFinished=TRUE;
                    }
                else
                  if (nRelation > 0)
                    if (pInternal->pGreater)
                      pInternal=pInternal->pGreater;
                    else
                      {
                        if (pExternal=(PatternNodePtr)
                         malloc(sizeof(struct PatternNode)))
                          {
                            pExternal->szPattern=szPattern;
                            pExternal->nDistinctLetters=nDistinctLetters;
                            pExternal->ulPlainTexts=(unsigned long) 0;
                            pExternal->pPlainTextHead=NULL;
                            pExternal->pLesser=NULL;
                            pExternal->pGreater=NULL;
                            pInternal->pGreater=pExternal;
                          }
                        else
                          {
                            *bErr=TRUE;
                            printf("Fatal error:  out of memory.\n");
                          }
                        pInternal=pExternal;
                        bFinished=TRUE;
                      }
                  else
                    {
                      free((void *) szPattern);
                      bFinished=TRUE;
                    }
              }
            while (! bFinished);
          }
        else
          if (pInternal=(PatternNodePtr) malloc(sizeof(struct PatternNode)))
            {
              pInternal->szPattern=szPattern;
              pInternal->nDistinctLetters=nDistinctLetters;
              pInternal->ulPlainTexts=(unsigned long) 0;
              pInternal->pPlainTextHead=NULL;
              pInternal->pLesser=NULL;
              pInternal->pGreater=NULL;
              *pPatternHead=pInternal;
            }
          else
            {
              *bErr=TRUE;
              printf("Fatal error:  out of memory.\n");
            }
      }
    return pInternal;
  }

static PatternNodePtr AddEncipheredWord(
 char              *szEncipheredWord,
 EncipheredNodePtr *pSortedHead,
 EncipheredNodePtr *pEncipheredTail,
 PatternNodePtr    *pPatternHead,
 int               *bErr)
  {
    EncipheredNodePtr pExternal;
    PatternNodePtr    pPattern;

    pPattern=NULL;
    if (pExternal=(EncipheredNodePtr) malloc(sizeof(struct EncipheredNode)))
      if (pExternal->szCipherText=(char *) malloc(1+strlen(szEncipheredWord)))
        {
          strcpy(pExternal->szCipherText,szEncipheredWord);
          pExternal->pSorted=NULL;
          pPattern=PatternForEncipheredWord(szEncipheredWord,pPatternHead,bErr);
          pExternal->pPattern=pPattern;
          pExternal->pNext=NULL;
          if (*pEncipheredTail)
            (*pEncipheredTail)->pNext=pExternal;
          else
            *pSortedHead=pExternal;
          *pEncipheredTail=pExternal;
        }
      else
        {
          *bErr=TRUE;
          printf("Fatal error:  out of memory.\n");
        }
    else
      {
        *bErr=TRUE;
        printf("Fatal error:  out of memory.\n");
      }
    return pPattern;
  }

int _main(
 void)
  {
    int               bAborted;
    int               bAlphabetic;
    int               bDuplicateWord;
    int               bErr;
    int               bLetterUsed [26];
    double            dDistinctLettersPerWord;
    double            dMaxDistinctLettersPerWord;
    int               nCurrentChar;
    int               nDistinctWords;
    int               nLetter;
    int               nMatchingLetters;
    int               nMaxMatchingLetters;
    int               nWord;
    int               nWordLen;
    char              *pc;
    char              *pcCipherText;
    CipherTextNodePtr pCipherText;
    CipherTextNodePtr pCipherTextHead;
    EncipheredNodePtr pEncipheredHead;
    EncipheredNodePtr pEnciphered1;
    EncipheredNodePtr pEnciphered2;
    EncipheredNodePtr pEncipheredForMax;
    EncipheredNodePtr pEncipheredTail;
    PatternNodePtr    pPattern;
    PatternNodePtr    pPatternHead;
    SortedNodePtr     pSorted;
    SortedNodePtr     pSortedHead;
    SortedNodePtr     pSortedTail;
    char              *szLowerCase;
    char              szPattern [256];
    char              szWord [256];

    bErr=FALSE;
    bAborted=FALSE;
    clrscr();
    pCipherTextHead=NULL;
    GetCipherText(&pCipherTextHead,&bErr);
    if (! bErr)
      {
        pCipherText=pCipherTextHead;
        pcCipherText=pCipherTextHead->szCipherText;
        printf("Getting patterns in enciphered text...\n");
        nDistinctWords=0;
        pEncipheredHead=NULL;
        pEncipheredTail=NULL;
        pPatternHead=NULL;
        do
          {
            while (((nCurrentChar=GetCipherChar(&pCipherText,&pcCipherText))
                      != EOF)
            &&     (nCurrentChar != (int) '\'')
            &&     (! islower(nCurrentChar=tolower(nCurrentChar))));
            if (nCurrentChar != EOF)
              {
                szWord[0]='\0';
                nWordLen=0;
                do
                  szWord[nWordLen++]=(char) nCurrentChar;
                while (((nCurrentChar=GetCipherChar(&pCipherText,&pcCipherText)) != EOF)
                &&     (nWordLen < 255)
                &&     ((nCurrentChar == '\'')
                     || (islower(nCurrentChar=tolower(nCurrentChar)))));
                szWord[nWordLen]='\0';
                pPattern=AddEncipheredWord(&szWord[0],&pEncipheredHead,
                 &pEncipheredTail,&pPatternHead,&bErr);
                if ((! bErr)
                &&  (pPattern->pPlainTextHead == NULL))
                  AddWordNotInDictionary(pPattern,&bErr);
                bAborted=Aborted();
              }
          }
        while ((! bErr) && (! bAborted) && (nCurrentChar != EOF));
        if ((! bErr) && (! bAborted))
          {
            if (pPatternHead == NULL)
              {
                bErr=TRUE;
                printf("Fatal error:  the cipher text contains no words.\n");
              }
          }
        if ((! bErr) && (! bAborted))
          {
            printf("Getting words matching the patterns...\n");
            printf("Press a key to abort.\n");
            nWord=0;
            while ((! bErr)
            &&     (! bAborted)
            &&     (g_pszWordList[nWord]))
              if (szLowerCase=(char *)
               malloc(1+strlen(g_pszWordList[nWord])))
                {
                  strcpy(szLowerCase,g_pszWordList[nWord]);
                  GetPatternForPlainTextWord(szLowerCase,
                   &szPattern[0],&bAlphabetic);
                  if (bAlphabetic)
                    AddPlainText(pPatternHead,&szPattern[0],
                     szLowerCase,&bErr);
                  free((void *) szLowerCase);
                  ++nWord;
                  bAborted=Aborted();
                }
              else
                {
                  bErr=TRUE;
                  printf(
                   "Fatal error:  memory was exhausted.\n");
                }
            if ((! bErr) && (! bAborted))
              {
                dMaxDistinctLettersPerWord=(double) 0;
                pEncipheredForMax=NULL;
                pEnciphered1=pEncipheredHead;
                while (pEnciphered1)
                  {
                    if (pEnciphered1->pPattern->ulPlainTexts)
                      {
                        dDistinctLettersPerWord
                         =((double) (pEnciphered1->pPattern->
                         nDistinctLetters))
                         /((double) (pEnciphered1->pPattern->
                         ulPlainTexts));
                        if (dDistinctLettersPerWord
                         > dMaxDistinctLettersPerWord)
                          {
                            dMaxDistinctLettersPerWord
                             =dDistinctLettersPerWord;
                            pEncipheredForMax=pEnciphered1;
                          }
                      }
                    pEnciphered1=pEnciphered1->pNext;
                  }
                if (dMaxDistinctLettersPerWord <= (double) 0)
                  {
                    bErr=TRUE;
                    printf("Fatal error:  no word in the "
                     "dictionary matched any pattern in the cipher text.\n");
                  }
                else
                  if (pSortedHead=(SortedNodePtr)
                   malloc(sizeof(struct SortedNode)))
                    {
                      pEncipheredForMax->pSorted=pSortedHead;
                      pSortedHead->szCipherText
                       =pEncipheredForMax->szCipherText;
                      pSortedHead->pPattern
                       =pEncipheredForMax->pPattern;
                      pSortedHead->pPlainText
                       =pEncipheredForMax->pPattern->pPlainTextHead;
                      pSortedHead->pPrevious=NULL;
                      pSortedHead->pNext=NULL;
                      pSortedTail=pSortedHead;
                      for (nLetter=26; nLetter--;)
                        bLetterUsed[nLetter]=FALSE;
                      pc=pEncipheredForMax->szCipherText;
                      while (*pc)
                        {
                          if (*pc != '\'')
                            bLetterUsed[tolower((int) *pc)-(int) 'a']
                             =TRUE;
                          ++pc;
                        }
                      pEnciphered1=pEncipheredHead->pNext;
                      while ((! bErr)
                      &&     (! bAborted)
                      &&     (pEnciphered1))
                        {
                          nMaxMatchingLetters=0;
                          pEncipheredForMax=NULL;
                          pEnciphered2=pEncipheredHead;
                          while (pEnciphered2)
                            {
                              if (pEnciphered2->pSorted == NULL)
                                {
                                  nMatchingLetters=0;
                                  pc=pEnciphered2->szCipherText;
                                  while (*pc)
                                    {
                                      if ((*pc != '\'')
                                      &&  (bLetterUsed[
                                            tolower((int) *pc)
                                            -(int) 'a']))
                                        ++nMatchingLetters;
                                      ++pc;
                                    }
                                  if (nMatchingLetters
                                   >= nMaxMatchingLetters)
                                    {
                                      nMaxMatchingLetters
                                       =nMatchingLetters;
                                      pEncipheredForMax=pEnciphered2;
                                    }
                                }
                              pEnciphered2=pEnciphered2->pNext;
                            }
                          bDuplicateWord=FALSE;
                          pEnciphered2=pEncipheredHead;
                          while ((! bDuplicateWord)
                          &&     (pEnciphered2))
                            if (pEnciphered2->pSorted)
                              if (MyStricmp(pEnciphered2->szCipherText,
                               pEncipheredForMax->szCipherText))
                                pEnciphered2=pEnciphered2->pNext;
                              else
                                bDuplicateWord=TRUE;
                            else
                              pEnciphered2=pEnciphered2->pNext;
                          if (bDuplicateWord)
                            pEncipheredForMax->pSorted
                             =pEnciphered2->pSorted;
                          else
                            if (pSorted=(SortedNodePtr)
                             malloc(sizeof(struct SortedNode)))
                              {
                                ++nDistinctWords;
                                pEncipheredForMax->pSorted=pSorted;
                                pSorted->szCipherText
                                 =pEncipheredForMax->szCipherText;
                                pSorted->pPattern
                                 =pEncipheredForMax->pPattern;
                                pSorted->pPlainText=pEncipheredForMax->
                                 pPattern->pPlainTextHead;
                                pSortedTail->pNext=pSorted;
                                pSorted->pPrevious=pSortedTail;
                                pSorted->pNext=NULL;
                                pSortedTail=pSorted;
                                pc=pEncipheredForMax->szCipherText;
                                while (*pc)
                                  {
                                    if (*pc != '\'')
                                     bLetterUsed[tolower((int) *pc)
                                      -(int) 'a']=TRUE;
                                    ++pc;
                                  }
                              }
                            else
                              {
                                bErr=TRUE;
                                printf(
                                 "Fatal error:  out of memory.\n");
                              }
                          pEnciphered1=pEnciphered1->pNext;
                          bAborted=Aborted();
                        }
                      if ((! bErr) && (! bAborted))
                        {
                          printf("Solving...\n");
                          Solve(nDistinctWords,pEncipheredHead,
                           pSortedHead,pCipherTextHead,&bAborted);
                        }
                      while (pSortedHead)
                        {
                          pSorted=pSortedHead->pNext;
                          free((void *) pSortedHead);
                          pSortedHead=pSorted;
                        }
                    }
                  else
                    {
                      bErr=TRUE;
                      printf("Fatal error:  out of memory.\n");
                    }
              }
          }
        FreePattern(pPatternHead);
        while (pEncipheredHead)
          {
            pEnciphered1=pEncipheredHead->pNext;
            free((void *) (pEncipheredHead->szCipherText));
            free((void *) pEncipheredHead);
            pEncipheredHead=pEnciphered1;
          }
      }
    while (pCipherTextHead)
      {
        pCipherText=pCipherTextHead->pNext;
        free((void *) (pCipherTextHead->szCipherText));
        free((void *) pCipherTextHead);
        pCipherTextHead=pCipherText;
      }
    printf("Press a key to exit.");
    ngetchx();
    return bErr;
  }
